@@ -31,4 +31,9 @@ typedef struct _GrdPipeWireStream GrdPipeWireStream;
 typedef struct _GrdPipeWireStreamMonitor GrdPipeWireStreamMonitor;
 typedef struct _GrdVncServer GrdVncServer;
 
+typedef enum _GrdPixelFormat
+{
+  GRD_PIXEL_FORMAT_RGBA8888,
+} GrdPixelFormat;
+
 #endif /* GRD_TYPES_H */
